import abc
import uuid


class TriggerInput(abc.ABC):
    @abc.abstractmethod
    def __init__(self, *args, **kwargs): ...
    @abc.abstractmethod
    async def receive_job_id(self) -> uuid.UUID: ...



import abc


class CameraOutput(abc.ABC):
    @abc.abstractmethod
    def __init__(self, *args, **kwargs): ...
    @abc.abstractmethod
    def write(self, buf: bytes) -> int: ...
    @abc.abstractmethod
    async def awrite(self, buf: bytes) -> int: ...



from pydantic import Field
from pydantic_settings import SettingsConfigDict

from .base import CfgBaseSettings


class CfgCameraVirtual(CfgBaseSettings):
    model_config = SettingsConfigDict(env_prefix="camera_virtual_")

    # server: str = Field(default="0.0.0.0")

    fps_nominal: int = Field(default=10)



import uuid

import pynng

from .base import TriggerInput


class PynngTriggerInput(TriggerInput):
    def __init__(self, address: str):
        self.__sub = pynng.Sub0()
        self.__sub.subscribe(b"")
        self.__sub.listen(address=address)

    async def receive_job_id(self) -> uuid.UUID:
        """Encapsulates arecv and converts to UUID."""
        msg = await self.__sub.arecv()
        return uuid.UUID(bytes=msg)



import abc
import uuid

from .output.base import CameraOutput


class CameraBackend(abc.ABC):
    @abc.abstractmethod
    def __init__(self, device_id: int, output_lores: CameraOutput, output_hires: CameraOutput):
        self._device_id = device_id
        self._output_lores = output_lores
        self._output_hires = output_hires

    @abc.abstractmethod
    async def run(self): ...
    @abc.abstractmethod
    async def trigger_hires_capture(self, job_id: uuid.UUID): ...



import asyncio
import os
import uuid

import cv2
import numpy as np
import pynng

from wigglecam.dto import ImageMessage

DEVICES = [
    ("localhost", 5550),  # connect to, base-port
    ("localhost", 5560),
]


async def main():
    # host also subscribes to the hires replies
    pub_trigger = pynng.Pub0()
    sub_lo = pynng.Sub0()  # Pipeline pull for lores stream.
    sub_lo.subscribe(b"")
    sub_hi = pynng.Sub0()  # Pipeline pull for lores stream.
    sub_hi.subscribe(b"")
    sub_hi.recv_timeout = 1000

    # def cb_connected(pipe: pynng.Pipe):
    #     print("Verbunden:", len(pub_trigger.pipes))

    # def cb_disconnected(pipe: pynng.Pipe):
    #     active = [p for p in pub_trigger.pipes if not p.closed]
    #     print("Noch verbunden:", len(active))
    #     print("Noch verbunden:", len(pub_trigger.pipes))

    # pub_trigger.add_post_pipe_connect_cb(cb_connected)
    # pub_trigger.add_post_pipe_remove_cb(cb_disconnected)

    # print(len(pub_trigger.dialers))

    for host, base_port in DEVICES:
        # block=False means continue program and try to connect in the background without any exception
        print(f"tcp://{host}:{base_port + 0}")
        pub_trigger.dial(f"tcp://{host}:{base_port + 0}", block=False)
        sub_lo.dial(f"tcp://{host}:{base_port + 1}", block=False)
        sub_hi.dial(f"tcp://{host}:{base_port + 2}", block=False)

    print(f"listen on base ports {[port[1] for port in DEVICES]} for devices")

    lores_frames = {}
    trigger = asyncio.Event()

    async def lores_task():
        nonlocal lores_frames
        while True:
            data = await sub_lo.arecv()
            msg = ImageMessage.from_bytes(data)

            img = cv2.imdecode(np.frombuffer(msg.jpg_bytes, np.uint8), cv2.IMREAD_COLOR)
            lores_frames[msg.device_id] = img

    async def hires_task():
        base_dir = "tmp/job_results"
        os.makedirs(base_dir, exist_ok=True)

        while True:
            await trigger.wait()
            trigger.clear()
            print("Job start")

            print("connected devices:", len(pub_trigger.pipes))
            if len(DEVICES) > len(pub_trigger.pipes):
                print("WARNING: not all devices are connected!")

            # Eindeutige ID für diese Umfrage
            job_uuid = uuid.uuid4()
            await pub_trigger.asend(job_uuid.bytes)

            job_folder = os.path.join(base_dir, f"job_{job_uuid}")
            os.makedirs(job_folder, exist_ok=True)
            results_no = 0

            while True:
                try:
                    data = await sub_hi.arecv()
                    msg = ImageMessage.from_bytes(data)
                    results_no += 1

                    if msg.job_id != job_uuid:
                        # Antwort gehört zu alter Umfrage -> ignorieren
                        print("warning, old job id result received, ignored!")
                        continue

                    fname = f"cam{msg.device_id}.jpg"
                    with open(os.path.join(job_folder, fname), "wb") as f:
                        f.write(msg.jpg_bytes)

                    if results_no == len(DEVICES):
                        print("got all results, job completed!")
                        break

                except pynng.exceptions.Timeout:
                    # raise RuntimeError("tis bad, we got not all results!")
                    print(f"job finished incomplete after 1s no more data, got {results_no} result but {len(DEVICES)} expected!")
                    break

    async def ui_task():
        nonlocal lores_frames
        while True:
            if lores_frames:
                imgs = [lores_frames[cid] for cid in sorted(lores_frames.keys())]
                h = 240
                imgs_resized = [cv2.resize(img, (320, h)) for img in imgs]
                rows = []
                for i in range(0, len(imgs_resized), 2):
                    row_imgs = imgs_resized[i : i + 2]
                    if len(row_imgs) == 1:
                        # pad with a black image of same size
                        blank = np.zeros_like(row_imgs[0])
                        row_imgs.append(blank)
                    row = cv2.hconcat(row_imgs)
                    rows.append(row)

                wall = cv2.vconcat(rows)
                cv2.imshow("Live Wall", wall)

            key = cv2.waitKey(1)
            if key == 27:  # ESC
                break
            elif key == ord("t"):
                trigger.set()

            await asyncio.sleep(0.05)

    await asyncio.gather(lores_task(), hires_task(), ui_task())


def run_async():
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        print("Exit app.")


if __name__ == "__main__":
    run_async()



from typing import Literal

from pydantic import Field
from pydantic_settings import SettingsConfigDict

from .base import CfgBaseSettings


class CfgCameraPicamera2(CfgBaseSettings):
    model_config = SettingsConfigDict(env_prefix="camera_picamera2_")

    camera_num: int = Field(default=0)
    framerate: int = Field(default=8)
    optimize_memoryconsumption: bool = Field(default=True)
    software_sync: Literal["off", "server", "client"] = Field(default="off")

    camera_res_width: int = Field(default=2304)  # max 2304 if HDR type imx708, otherwise 4608
    camera_res_height: int = Field(default=1296)  # max 1296 if HDR type imx708, otherwise 2592
    hdr_type: Literal["unset", "imx708", "pi5"] = Field(
        default="unset",
        description="Choose whatever hardware is used. Unset does not actively disable, but just do nothing. To disable, unpower.",
    )

    stream_res_width: int = Field(default=1152)
    stream_res_height: int = Field(default=648)
    frame_skip_count: int = Field(
        default=2,
        ge=1,
        le=4,
        description="Reduce the stream framerate by frame_skip_count to save cpu/gpu/network on producing device as well as client devices. Choose 1 to emit every produced frame.",
    )

    flip_horizontal: bool = Field(default=False)
    flip_vertical: bool = Field(default=False)

    videostream_quality: Literal["VERY_LOW", "LOW", "MEDIUM", "HIGH", "VERY_HIGH"] = Field(default="HIGH")



from pydantic_settings import BaseSettings, SettingsConfigDict


class CfgBaseSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_file_encoding="utf-8",
        # first in following list is least important; last .env file overwrites the other.
        env_file=[".env.dev", ".env.test", ".env.primary", ".env.node"],
        env_nested_delimiter="__",
        case_sensitive=False,
        extra="ignore",
    )
